
import { Player, system, world, Direction, EquipmentSlot, ItemStack, MolangVariableMap, World } from "@minecraft/server";
// file is saved as scripts/minecraft-math.js
import { Vector3Builder } from "minecraft-math.js";
import { Vector3Utils } from "./minecraft-math";
// file is saved as scripts/vanilla-data.js
import { MinecraftBlockTypes } from "vanilla-data.js";
import { MinecraftEffectTypes } from "vanilla-data.js";
import { MinecraftEntityTypes } from "vanilla-data.js";



function print(msg) {
    world.sendMessage(msg);
}

// // Diamond chicken buffs
// world.afterEvents.effectAdd.subscribe((data) =>{
//     const player = data.entity;
//     const effect = data.effect;
// 
//     if(player.typeId == MinecraftEntityTypes.Player && player.hasTag("yeggs_chicken:diamond_buff") && effect.amplifier == 1 && effect.duration == 30*20 && (effect.typeId == "speed" || effect.typeId == "strength" ||effect.typeId == "regeneration"))
//     {
//         print("buff");
//         const effectId = effect.typeId;
//         
//         const callback = system.runInterval(() => {
//             if(player.isValid())
//             {
//                 const hasEffect = player.getEffect(effectId);
// 
//                 // Loop until they lose effect   
//                 if(!hasEffect)
//                 {
//                     print("end");
// 
//                     const scoreboardIdentity = player.scoreboardIdentity;
//                     let objective = world.scoreboard.getObjective("yeggs_chicken:buff_count");
//                     if (!objective) {
//                         objective = world.scoreboard.addObjective("yeggs_chicken:buff_count");
//                     }
// 
//                     let playerScore = objective.getScore(scoreboardIdentity) ?? 0;
// 
//                     objective.setScore(scoreboardIdentity, playerScore-1);
// 
//                     system.clearRun(callback);
//                 }
//             }
//         }, 0)
//     }
// })


world.afterEvents.entityHitBlock.subscribe((data) =>{

    const player = data.damagingEntity;
    const equippable = player.getComponent("minecraft:equippable");
    const item = equippable.getEquipment("Mainhand");
    const dimension = player.dimension;
    

    if (item && item.typeId === "yeggs_chicken:chicken_wand") {

        let cooldown = item.getComponent("minecraft:cooldown");
        let time = cooldown.getCooldownTicksRemaining(player)

        if(time > 0)
            return;

        let equippable = player.getComponent("minecraft:equippable");
        let inventory = player.getComponent("minecraft:inventory").container;
        let shardIndex;
        let shardStack;

        let hasShard = false;
        for (let i = 0; i < inventory.size; i++) {
            let curItem = inventory.getItem(i);
            if (curItem && curItem.typeId === "yeggs_chicken:chicken_shard")
            {
                shardIndex = i;
                shardStack = curItem;
                hasShard = true;
            }
        }

        if (!hasShard)
        {
            dimension.playSound("note.didgeridoo", player.location, { volume: 1, pitch: 1.5 })
            player.sendMessage("§cYou have no Chicken Shards.")
            return;
        }


        if(shardStack.amount <= 1)
            inventory.setItem(shardIndex, undefined)
        else
        {
            shardStack.amount--
            inventory.setItem(shardIndex, shardStack)
        }

        let hitDirection = data.blockFace;
        let hitPosition = data.hitBlock.location;
        switch(hitDirection)
        {
            case "Down":
                hitPosition = Vector3Utils.add(hitPosition,{x:0,y:-1,z:0});
            case "Up":
                hitPosition = Vector3Utils.add(hitPosition,{x:0,y:1,z:0});
            case "Ease":
                hitPosition = Vector3Utils.add(hitPosition,{x:1,y:0,z:0});
            case "West":
                hitPosition = Vector3Utils.add(hitPosition,{x:-1,y:0,z:0});
            case "North":
                hitPosition = Vector3Utils.add(hitPosition,{x:0,y:0,z:1});
            case "South":
                hitPosition = Vector3Utils.add(hitPosition,{x:0,y:0,z:-1});
        }

        cooldown.startCooldown(player);

        dimension.playSound("mob.evocation_illager.cast_spell", hitPosition, { volume: 1, pitch: 1.5 })
        
        dimension.spawnEntity("minecraft:chicken", Vector3Utils.add(hitPosition,{x:0.5,y:0,z:0.5}));

    }
})

world.afterEvents.itemUse.subscribe((data) => {
    const item = data.itemStack;
    const player = data.source;
    const dimension = player.dimension;

    if (item.typeId === "yeggs_chicken:chicken_wand") {
        let cooldown = item.getComponent("minecraft:cooldown");
        let time = cooldown.getCooldownTicksRemaining(player)

        let equippable = player.getComponent("minecraft:equippable");

        if(player.getDynamicProperty("yeggs_chicken:chicken_selected"))
        {
            player.setDynamicProperty("yeggs_chicken:chicken_shoot", true)
            player.setDynamicProperty("yeggs_chicken:chicken_selected", false)
        }
        else
        {
            if (time == 0) {

                if (!(player instanceof Player)) return;
    
                chickenWandSelect(item, player, cooldown);
    
                equippable.setEquipment(EquipmentSlot.Mainhand, item);
            }
        }
    }
})

function chickenWandSelect(item, player, cooldown) {
    const dimension = player.dimension;
    let maxDistance = 5;
    const dotThreshold = 0.1;

    // Check if player is looking at chicken
    //const chickens = player.getEntitiesFromViewDirection({ maxDistance: maxDistance, families: ["chicken"] });
    let chicken;

    const origin = player.location;
    const targets = dimension.getEntities({ location: origin, maxDistance: maxDistance, families:["chicken"]});
    const playerDirection = player.getViewDirection();
    // Loop through all nearby entities (not items though)
    targets.forEach(target => {
        // Dont target shooting player
        if (target == player) return;

        // Calculate dot product
        let targetDirection = new Vector3Builder(0, 0, 0);
        targetDirection = Vector3Utils.normalize(Vector3Utils.subtract(target.location, player.location));
        let targetDotProduct = Vector3Utils.dot(playerDirection, targetDirection);

        if (targetDotProduct >= 1 - dotThreshold) {
            chicken = target;
        }
    });

    if (chicken) {
        cooldown.startCooldown(player);

        player.setDynamicProperty("yeggs_chicken:chicken_selected", true);


        chicken.setDynamicProperty("yeggs_chicken:chicken_selected", true);

        dimension.playSound("beacon.activate", player.location, { volume: 1, pitch: 3 })

        const callback = system.runInterval(() => {


            // Check if the chicken dies/despawns
            if(!chicken.isValid())
            {
                player.setDynamicProperty("yeggs_chicken:chicken_selected", false);
                player.setDynamicProperty("yeggs_chicken:chicken_shoot", false);
                system.clearRun(callback);
                return;
            }

            // If the player has shot
            if(player.isValid() && player.getDynamicProperty("yeggs_chicken:chicken_shoot"))
                {
                    chickenWandShoot(player,chicken, player.getViewDirection(), 5);
                    player.setDynamicProperty("yeggs_chicken:chicken_shoot", false);
                }
                


            // Check player is still holding wand
            if(player.isValid())
            {
                const playerItem = player.getComponent('equippable').getEquipment('Mainhand');
                if(playerItem.typeId !== "yeggs_chicken:chicken_wand")
                {
                    player.setDynamicProperty("yeggs_chicken:chicken_selected", false);
                }
            }

            // Check if the player leaves or has stopped holding wand
            if (!player.isValid() || !player.getDynamicProperty("yeggs_chicken:chicken_selected")) {
                chicken.setDynamicProperty("yeggs_chicken:chicken_selected", false);
                system.clearRun(callback);
                return;
            }
            else
            {
                    chickenWandMoveChicken(player, chicken, maxDistance);
            }

            player.setDynamicProperty("yeggs_chicken:chicken_shoot", false);

        }, 0)

    }
}

function chickenWandMoveChicken(player, chicken, maxDistance)
{
    let dimension = player.dimension;
    let playerDirection = player.getViewDirection();

    let isValid = true;
    let currentDistance = 0.0;
    let currentPosition = player.getHeadLocation();

    while (isValid && currentDistance < maxDistance - 1) {
        currentDistance+=0.5;

        currentPosition = Vector3Utils.add(currentPosition, Vector3Utils.scale(playerDirection,0.5));

        // If there is a block in the current position
        let checkBlock = dimension.getBlockFromRay(Vector3Utils.subtract(currentPosition, Vector3Utils.scale(playerDirection, 0.499)), playerDirection, { maxDistance: 1, includeLiquidBlocks: false, includePassableBlocks: false });
        if (checkBlock || currentDistance >= maxDistance-1) {
            chicken.teleport( Vector3Utils.subtract(Vector3Utils.subtract(currentPosition, playerDirection),{x:0,y:0.25,z:0}), {facingDirection:Vector3Utils.scale(playerDirection, -1), keepVelocity:false});
            isValid = false;
        }

    }
}

function chickenWandShoot(player,chicken,direction, strength)
{

    player.dimension.playSound("beacon.power", player.location, { volume: 1, pitch: 3 })

    const dimension = player.dimension;

    chicken.setDynamicProperty("yeggs_chicken:chicken_selected", false);
    player.setDynamicProperty("yeggs_chicken:chicken_selected", false);

    chicken.applyImpulse(Vector3Utils.scale(direction, strength));

    let hitLength = 0;
    let hitLengthMax = 2;
    let hitDelay = 0;
    let hitDelayMax = 5;
    let entitiesHit = [];

    // Until chicken hits another chicken or block
    const callback = system.runInterval(() => {

        hitDelay++;

        // Check if the chicken dies/despawns
        if (!chicken.isValid()) {
            system.clearRun(callback);
            return;
        }

        chicken.applyImpulse({ x: 0, y: -0.1, z: 0 });

        if(hitDelay < hitDelayMax) return;
        // Check collision with ground below and infront
        let checkHitFloor = dimension.getBlockFromRay(chicken.location, { x: 0, y: -1, z: 0 }, { maxDistance: 1, includeLiquidBlocks: true, includePassableBlocks: false });
        let checkHitWall = dimension.getBlockFromRay(chicken.location, direction, { maxDistance: 1, includeLiquidBlocks: true, includePassableBlocks: false });
        if (checkHitFloor || checkHitWall) 
            hitLength++;
        else
            hitLength = 0;

        if (hitLength > hitLengthMax) {
            // EXPLODE
            dimension.spawnParticle('minecraft:huge_explosion_emitter', chicken.location);
            dimension.playSound("random.explode", chicken.location, { volume: 3, pitch: 0.75 });
            const entities = dimension.getEntities({location: chicken.location,  maxDistance: 5, excludeFamilies: ["inanimate"], excludeTypes: ["item"] })

            //if(entities.length > 0)
            {
                entities.forEach(target => {
                    // Dont target shooting player or chicken
                    if (target == player || target == chicken) return;

                    target.applyDamage(4, { damagingEntity: player, damagingProjectile: chicken });
                });
            }

            system.clearRun(callback);
            return;
        }

        //// Get check volume
        //let volumeX = 1;
        //let volumeY = 1;
        //let volumeZ = 1;
//
        //// Check collision with another entity
        //const entities1 = dimension.getEntities({ location: Vector3Utils.subtract(chicken.location,chicken.getVelocity()), volume: { x: volumeX, y: volumeY, z: volumeZ }, excludeFamilies: ["inanimate"], excludeTypes: ["item"] });
        //const entities2 = dimension.getEntities({ location: chicken.location, volume: { x: volumeX, y: volumeY, z: volumeZ }, excludeFamilies: ["inanimate"], excludeTypes: ["item"] });
        //const entities3 = dimension.getEntities({ location: Vector3Utils.add(chicken.location,direction), volume: { x: volumeX, y: volumeY, z: volumeZ }, excludeFamilies: ["inanimate"], excludeTypes: ["item"] });
        //const entities = entities1.concat(entities2).concat(entities3);
        //// Loop through all hit entities
        //if(entities.length > 0)
        //{
        //    entities.forEach(target => {
        //        // Dont target shooting player or chicken
        //        if (target == player || target == chicken) return;
    //
        //        if (entitiesHit.includes(target)) return;
    //
        //        entitiesHit.push(target);
    //
        //        target.applyDamage(4, { damagingEntity: player, damagingProjectile: chicken });
        //    });
        //}

    }, 0)

}





/**
 * @param {Player} player
 * @param {Entity} projectileType
 * @param {number?} power 
 */
function shootProjectile(player, projectileType, power) {

    const view = player.getViewDirection(),
        head = player.getHeadLocation(),
        direction = { x: head.x + view.x, y: head.y + view.y, z: head.z + view.z },
        velocity = { x: view.x * power, y: view.y * power, z: view.z * power };
    const projectile = player.dimension.spawnEntity(projectileType, direction).getComponent("minecraft:projectile");
    projectile.shoot(velocity)
    projectile.owner = player
};
